<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 3/13/17
 * Time: 9:29 PM
 */

namespace qfq;

use qfq;

require_once(__DIR__ . '/../store/Session.php');
require_once(__DIR__ . '/../Constants.php');
require_once(__DIR__ . '/../database/Database.php');
require_once(__DIR__ . '/../../qfq/store/Client.php');

/**
 * Class Dirty
 * Process Record locking in mode DIRTY_MODE_ADVISORY, DIRTY_MODE_EXCLUSIVE or DIRTY_MODE_NONE.
 * Two entry points: process() and checkDirtyAndRelease().
 * Check doc/diagram/*.png for detailed workflow.
 *
 * @package qfq
 */
class Dirty {

    /**
     * @var Database instantiated class
     */
    protected $db = null;

    /**
     * @var array
     */
    protected $client = array();

    /**
     * @var Session
     */
    private $session = null;

    /**
     * @var array()
     */
    private $lockTimeout = false;

    /**
     * Init class
     */
    public function __construct($lockTimeout = false, $phpUnit = false) {

        $this->session = Session::getInstance($phpUnit);
        $this->client = Client::getParam();
        $this->db = new Database();
        if($lockTimeout === false) {
            $cfg = new Config();
            $config = $cfg->readConfig('');
            $this->lockTimeout = $config[SYSTEM_RECORD_LOCK_TIMEOUT_SECONDS];
        } else {
            $this->lockTimeout = $lockTimeout;
        }
    }

    /**
     * Handle any lock requests submitted via api/dirty.php.
     *
     * @return array|int
     * @throws CodeException
     * @throws DbException
     * @throws UserFormException
     */
    public function process() {
        $tableVars = array();

        $sipClass = new Sip();

        $sipVars = $sipClass->getVarsFromSip($this->client[SIP_SIP]);

        if (empty($sipVars[SIP_FORM])) {
            throw new CodeException("Missing 'form' in SIP. There might be something broken.", ERROR_DIRTY_MISSING_FORM_IN_SIP);
        }

        $recordId = empty($sipVars[SIP_RECORD_ID]) ? 0 : $sipVars[SIP_RECORD_ID];
        if($recordId>0) {
            $tableVars = $this->db->sql("SELECT tableName, dirtyMode FROM Form AS f WHERE f.name=?", ROW_EXPECT_1, [$sipVars[SIP_FORM]], "Form not found: '" . $sipVars[SIP_FORM] . "'");
        }

        switch($this->client[API_LOCK_ACTION]) {
            case API_LOCK_ACTION_LOCK:
                $answer = $this->acquireDirty($recordId, $tableVars);
                break;
            case API_LOCK_ACTION_RELEASE:
                $answer = $this->checkDirtyAndRelease(FORM_SAVE, $tableVars[F_DIRTY_MODE], $tableVars[F_TABLE_NAME], $recordId);
                break;
            default;
                throw new CodeException("Unknown action: " . $this->client[API_LOCK_ACTION], ERROR_DIRTY_UNKNOWN_ACTION);
        }

        return $answer;

    }

    /**
     * Tries to get a 'DirtyRecord'. Returns an array about success or failure.
     *
     * @param int $recordId
     * @param array $tableVars
     * @return array
     * @throws CodeException
     * @throws DbException
     * @internal param array $sipVars
     */
    private function acquireDirty($recordId, array $tableVars) {

        // For r=0 (new) , 'dirty' will always succeed.
        if ($recordId == 0) {
            return [API_STATUS => 'success', API_MESSAGE => ''];
        }

        $tableName = $tableVars[F_TABLE_NAME];
        $formDirtyMode = $tableVars[F_DIRTY_MODE];

        $feUser = $this->session->get(SESSION_FE_USER);

        // Look for already existing dirty record.
        $recordDirty = $this->getRecordDirty($tableName, $recordId);

        // Check if the record is timed out
        if (count($recordDirty) != 0 && $this->lockTimeout > 0) {
            $datetimeExpire = date_add(date_create($recordDirty[COLUMN_MODIFIED]), date_interval_create_from_date_string($this->lockTimeout . " second"));
            if ($datetimeExpire <= date_create("now")) {
                $this->deleteDirtyRecord($recordDirty[COLUMN_ID]);
                $recordDirty=array();
            }
        }

        if (count($recordDirty) == 0) {
            // No dirty record found.
            $answer = $this->writeDirty($this->client[SIP_SIP], $recordId, $tableName, $formDirtyMode, $feUser);
        } else {
            $answer = $this->conflict($recordDirty, $formDirtyMode);
        }

        return $answer;
    }

    /**
     * Load (if exist) a DirtyRecord (lock).
     *
     * @param $tableName
     * @param $recordId
     * @return array   DirtyRecord or empty array.
     * @throws CodeException
     * @throws DbException
     */
    private function getRecordDirty($tableName, $recordId) {

        $recordDirty = $this->db->sql("SELECT * FROM Dirty AS d WHERE d.tableName LIKE ? AND recordId=? ",
            ROW_EXPECT_0_1, [$tableName, $recordId]);

        return $recordDirty;

    }

    /**
     *
     * @param array $recordDirty
     * @param $currentFormDirtyMode
     * @return array
     */
    private function conflict(array $recordDirty, $currentFormDirtyMode) {

        $status = API_ANSWER_STATUS_CONFLICT;
        $at = "at " . $recordDirty[COLUMN_CREATED] . " from " . $recordDirty[DIRTY_REMOTE_ADDRESS];

        if ($this->client[CLIENT_COOKIE_QFQ] == $recordDirty[DIRTY_QFQ_USER_SESSION_COOKIE]) {
            $msg = "The record has already been locked by you (maybe in another browser tab) $at!";
            $status = API_ANSWER_STATUS_CONFLICT_ALLOW_FORCE;

        } else {

            if (empty($recordDirty[DIRTY_FE_USER])) {
                $msgUser = "another user";
            } else {
                $msgUser = "user '$recordDirty[DIRTY_FE_USER]'";
            }

            $msg = "The record has already been locked by $msgUser at $at.";

            // Mandatory lock on Record or current Form?
            if ($recordDirty[F_DIRTY_MODE] == DIRTY_MODE_EXCLUSIVE || $currentFormDirtyMode == DIRTY_MODE_EXCLUSIVE) {
                $status = API_ANSWER_STATUS_CONFLICT;
            } else {
                $status = API_ANSWER_STATUS_CONFLICT_ALLOW_FORCE;
            }
        }

        return [API_STATUS => $status, API_MESSAGE => $msg];
    }

    /**
     * Write a 'Dirty'-Record.
     *
     * @param string $s SIP given by URL GET
     * @param int $recordId  extracted from SIP
     * @param string $tableName  extracted from SIP
     * @param string $formDirtyMode
     * @param string $feUser
     * @return array
     * @throws CodeException
     * @throws DbException
     */
    private function writeDirty($s, $recordId, $tableName, $formDirtyMode, $feUser) {

        $record = $this->db->sql("SELECT * FROM $tableName WHERE id=?", ROW_EXPECT_1, [$recordId], "Record to tag 'dirty' not found.");

        // Not all tables does have a modified column.
        $recordModified = empty($record[COLUMN_MODIFIED]) ? 0 : $record[COLUMN_MODIFIED];

        // Write 'dirty' record
        $this->db->sql("INSERT INTO Dirty (`sip`, `tableName`, `recordId`, `recordModified`, `feUser`, `qfqUserSessionCookie`, `dirtyMode`, `remoteAddress`, `created`) " .
            "VALUES ( ?,?,?,?,?,?,?,?,? )", ROW_REGULAR, [$s, $tableName, $recordId, $recordModified, $feUser, $this->client[CLIENT_COOKIE_QFQ], $formDirtyMode, $this->client[CLIENT_REMOTE_ADDRESS], date('YmdHis')]);

        return [API_STATUS => API_ANSWER_STATUS_SUCCESS, API_MESSAGE => '', API_LOCK_TIMEOUT => $this->lockTimeout ];

    }

    /**
     * Release a dirtyRecord. This is only possible if the current user owns the dirtyRecord.
     * In case of not owner, throws an exception and the save should break.
     *
     * @param string $formMode  FORM_DELETE, FORM_SAVE
     * @param string $dirtyMode   DIRTY_MODE_EXCLUSIVE, DIRTY_MODE_ADVISORY, DIRTY_MODE_NONE
     * @param string $tableName
     * @param int $recordId
     * @return array
     * @throws CodeException
     * @throws UserFormException
     */
    public function checkDirtyAndRelease($formMode, $dirtyMode, $tableName, $recordId) {

        $answer = [ API_STATUS => API_ANSWER_STATUS_SUCCESS, API_MESSAGE => ''];

        if ($recordId == 0) {
            return $answer; // New records never have a recordDirty nor a conflict.
        }

        $recordDirty = $this->getRecordDirty($tableName, $recordId);

        if (empty($recordDirty) && $dirtyMode == DIRTY_MODE_NONE) {
            return $answer; // only situation where it's ok that there is no dirtyRecord.
        }

        if (empty($recordDirty)) {

            if ($formMode == FORM_DELETE) {
                return $answer;
            }

            throw new UserFormException("Missing record lock: please reload the form, edit and save again.");
        }

        $msgUser = (empty($recordDirty[DIRTY_FE_USER])) ? "another user" : "user '$recordDirty[DIRTY_FE_USER]'";
        $msgAt = "at " . $recordDirty[COLUMN_CREATED] . " from " . $recordDirty[DIRTY_REMOTE_ADDRESS];

        if ($formMode == FORM_DELETE) {
            throw new UserFormException("The record has been locked by $msgUser $msgAt");
        }

        // Is the dirtyRecord mine?
        if ($recordDirty[DIRTY_QFQ_USER_SESSION_COOKIE] == $this->client[CLIENT_COOKIE_QFQ]) {

            //TODO Check ob der Record seit dem Lock veraendert wurde. Falls ja, eine Meldung und Frage was zu tun ist.
            // Clear the lock
            $this->deleteDirtyRecord($recordDirty[COLUMN_ID]);
            return $answer;
        }

        //----------------------------------------
        // From here: there is a foreign lock!

        // Check if overwrite is allowed
        if ($dirtyMode == DIRTY_MODE_ADVISORY && $recordDirty[F_DIRTY_MODE] == DIRTY_MODE_ADVISORY) {
            return $answer;
        }

        // Check if the record is timed out
        if ($this->lockTimeout > 0) {
            $datetimeExpire = date_add(date_create($recordDirty[COLUMN_MODIFIED]), date_interval_create_from_date_string($this->lockTimeout . " second"));
            if ($datetimeExpire <= date_create("now")) {
                $this->deleteDirtyRecord($recordDirty[COLUMN_ID]);
                return $answer;
            }
        }

        throw new UserFormException("Save not possible - the record has been locked by $msgUser $msgAt");
    }

    /**
     * Delete the dirtyRecord with $recordDirtyId. Throw an exception if the record has not been deleted.
     *
     * @param $recordDirtyId
     * @throws CodeException
     * @throws DbException
     */
    private function deleteDirtyRecord($recordDirtyId) {
        $cnt = $this->db->sql('DELETE FROM Dirty WHERE id=? LIMIT 1', ROW_REGULAR, [$recordDirtyId]);
        if ($cnt != 1) {
            throw new CodeException("Failed to delete dirty record id=" . $recordDirtyId, ERROR_DIRTY_DELETE_RECORD);
        }
    }
}